from robotooter import RoboTooter
from robotooter.bots.base_bot import BaseBot
from robotooter.cli.util import get_yes
from robotooter.mastodon.authenticator import Authenticator


def run_authorize(rt: RoboTooter, bot: BaseBot, force: bool=False) -> None:
    if not bot.config.username:
        raise RuntimeError("Username not set")
    authenticator = Authenticator(
        main_path=rt.config.working_directory,
        bot_path=bot.config.working_directory,
        full_username=bot.config.username,
    )

    if authenticator.bot_config_exists:
        if force or get_yes("Mastodon authentication file already exists. Do you want to replace it?"):
            authenticator.bot_config_path.unlink()
        else:
            print("Keeping current authentication settings.")
            return

    print("\nGreat! Now we will request an authorization URL.")
    auth_url = authenticator.get_auth_url()
    print("Here's your authorization URL:")
    print(auth_url)

    print("\nYou need to open that in a browser, make sure you're logged into Mastodon, and authorize")
    print("the application to user the account. It will give you a code, which you will enter here.")
    code = input("Enter authorization code: ")

    authenticator.log_in(code=code)

    print(f"Success! {bot.config.bot_name} should be set to toot up a storm.")



import re
import shutil
from importlib import resources as resources
from pathlib import Path
from typing import Protocol

RE_BOTLESS = re.compile('Bot$')
RE_MODNAME = re.compile(r'(?<!^)(?=[A-Z])')


class HasWorkingDirectory(Protocol):
    working_directory: Path

    def pkg_resources_path(self) -> Path: ...

    def resource_path(self, sub_path: str | None = None) -> Path: ...

class ResourcesMixin:
    def pkg_resources_path(self) -> Path:
        botless = RE_BOTLESS.sub('', self.__class__.__name__)
        mod_name = RE_MODNAME.sub('_', botless).lower()
        return Path(str(resources.files(f"{mod_name}"))) / "resources"

    def install_package_resources(self: HasWorkingDirectory) -> None:
        shutil.copytree(self.pkg_resources_path(), self.resource_path())

    def resource_path(self: HasWorkingDirectory, sub_path: str | None = None) -> Path:
        p = self.working_directory / "resources"
        if sub_path is not None:
            return p / sub_path
        return p



from robotooter.filters.base_filter import BaseFilter


class BlankLineFilter(BaseFilter):
    def process(self, line: str) -> str | None:
        if line.strip() == "":
            return None
        return line

    def finalize(self) -> str | None:
        return ''

    def reset(self) -> None:
        pass



from pathlib import Path
from typing import Any

from robotooter.models.statuses import RtMedia as RtMedia
from robotooter.models.statuses import RtStatus as RtStatus
from robotooter.models.statuses import StatusVisibility as StatusVisibility

class FakeClient:
    status_calls: list[dict[str, Any]]
    media_calls: list[str]
    login_calls: list[list[str]]
    auth_url_call_count: int
    def __init__(self, output_path: Path) -> None: ...
    def status_post(self,
        status: str,
        in_reply_to_id: str | None = None,
        media_ids: list[str] | None = None,
        spoiler_text: str | None = None,
        quote_id: str | None = None,
        visibility: StatusVisibility | None = None
    ) -> RtStatus: ...
    def media_post(self, media_file: str, description: str) -> RtMedia: ...
    def get_auth_url(self) -> str: ...
    def log_in(self, code: str, to_file: str) -> None: ...



import argparse

from robotooter.bots.base_bot import COUNT_PARAM, BaseBot, command
from robotooter.filters.base_filter import BaseFilter
from robotooter.models.configs import BotConfig


class FakeBotConfig(BotConfig):
    extra_field_1: str = ''


class FakeBot(BaseBot[FakeBotConfig]):
    NAME = "FakeBot"
    DESCRIPTION = "FakeBot bot"

    CONFIG_CLASS = FakeBotConfig

    @classmethod
    def create_with_config(cls, config_data: BotConfig, filters: list[BaseFilter]) -> 'FakeBot':
        config = FakeBotConfig(**config_data.model_dump())
        return cls(config, filters)

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("-v1", "--value_one", type=float, default=1.0)
        parser.add_argument("-v2", "--value_two", type=str)

    def __init__(self, config: FakeBotConfig, filters: list[BaseFilter]):
        super().__init__(config, filters)
        self.number = 1

        self.command_overrides['toot'] = [
            ('-bk', '--book', {'type': str, 'help': 'Book'}),
            ('-cp', '--chapter', {'type': str, 'help': 'Chapter'}),
        ]

    def _setup_data(self, args: argparse.Namespace) -> None:
        assert args.value_one
        assert args.value_two

    def _toot(self, args: argparse.Namespace) -> None:
        assert args.chapter
        assert args.book

    @command(name='add_one', help_text='This is the custom help text.')
    def add_one(self, args: argparse.Namespace) -> None:
        self.number += 1

    @command(name='set_number', help_text='Set the number of the bot.', params=[COUNT_PARAM])
    def set_number(self, args: argparse.Namespace) -> None:
        self.number = args.count



from pathlib import Path

from _typeshed import Incomplete

from robotooter.file_filter import FileFilter as FileFilter
from robotooter.filters.base_filter import BaseFilter as BaseFilter

DEFAULT_STATE_SIZE: int

class MarkovProcessor:
    filters: Incomplete
    working_directory: Incomplete
    data_dir: Incomplete
    source_dir: Incomplete
    consolidated_file_path: Incomplete
    model_path: Incomplete
    def __init__(self, working_directory: Path, filters: list[BaseFilter]) -> None: ...
    def preprocess_sources(self) -> None: ...
    def build_model(self, state_size: int = ...) -> None: ...
    @property
    def consolidated_exists(self) -> bool: ...
    @property
    def model_exists(self) -> bool: ...



import importlib
import inspect
import os
import pkgutil
from abc import ABCMeta
from pathlib import Path
from typing import Type, TypeVar, Union

import requests

T = TypeVar('T')


def load_included(package_path: str, expected_klass: Type, ending: str) -> dict[str, Type]:
    included = {}

    # Import the main package
    package = importlib.import_module(package_path)

    # Get all submodules in the package
    for importer, modname, ispkg in pkgutil.iter_modules(package.__path__, package.__name__ + "."):
        try:
            # Import each submodule
            submodule = importlib.import_module(modname)

            # Look for filter classes in this submodule
            for name, obj in inspect.getmembers(submodule, inspect.isclass):
                if (issubclass(obj, expected_klass) and
                        obj is not expected_klass and
                        name.endswith(ending)):
                    included[name] = obj

        except ImportError as e:
            print(f"Failed to import {modname}: {e}")
            continue

    return included


def load_list(names: list[str], expected_type: Union[Type[T], ABCMeta]) -> dict[str, Type[T]]:
    klasses = {}
    for name in names:
        components = name.split('.')
        # Import the module containing the class
        module_path = '.'.join(components[:-1])
        class_name = components[-1]

        mod = importlib.import_module(module_path)
        klass = getattr(mod, class_name)
        if not issubclass(klass, expected_type):
            raise TypeError(f"{klass} is not a {expected_type.__name__}")
        klasses[klass.__name__] = klass
    return klasses


def download_sources(source_file_path: Path, output_directory: Path) -> None:
    if not os.path.exists(source_file_path):
        raise FileNotFoundError(f"{source_file_path} not found")

    if not os.path.exists(output_directory):
        os.makedirs(output_directory, exist_ok=True)

    with open(source_file_path, "r") as source_file:
        for line in source_file:
            url = line.strip()
            try:
                response = requests.get(url)
                output_name = url.split("/")[-1]
                with open(os.path.join(output_directory, output_name), "w") as output_file:
                    output_file.write(response.text)
            except Exception as e:
                print(f"Failed to download {url}: {e}")



from pathlib import Path
from typing import cast


class MastodonMixin:
    def split_username(self, full_username: str) -> tuple[str, str]:
        # if it starts with an @, remove that
        if full_username[0] == '@':
            full_username = full_username[1:]

        username, server = full_username.split('@')
        return username, server

    @property
    def bot_config_exists(self) -> bool:
        return self.bot_config_path.exists()

    @property
    def bot_config_path(self) -> Path:
        bot_path = cast(Path, getattr(self, 'bot_path'))
        return bot_path / 'mastodon.app'



import abc
from abc import ABC, abstractmethod

class BaseFilter(ABC, metaclass=abc.ABCMeta):
    def process(self, line: str) -> str | None: ...
    def finalize(self) -> str | None: ...
    @abstractmethod
    def reset(self) -> None: ...



from enum import StrEnum
from pathlib import Path

from pydantic import BaseModel

class RtMedia(BaseModel):
    id: str | None
    file_path: Path
    description: str | None

class StatusVisibility(StrEnum):
    Direct = 'direct'
    Private = 'private'
    Unlisted = 'unlisted'
    Public = 'public'

class RtStatus(BaseModel):
    id: str | None
    text: str
    in_reply_to_id: str | None
    url: str | None
    media: list[RtMedia]
    spoiler_text: str | None
    quote_id: str | None
    visibility: str | None



from pathlib import Path
from typing import Any

from robotooter.models.statuses import RtMedia, RtStatus, StatusVisibility


class FakeClient:
    def __init__(self, output_path: Path):
        self.status_calls: list[dict[str, Any]] = []
        self.media_calls: list[str] = []
        self._output_path = output_path
        self.login_calls: list[list[str]] = []
        self.auth_url_call_count = 0

    def status_post(
            self,
            status: str,
            in_reply_to_id: str | None = None,
            media_ids: list[str]|None = None,
            spoiler_text: str|None=None,
            quote_id: str|None=None,
            visibility: StatusVisibility|None=None
    ) -> RtStatus:
        current_call_id = len(self.status_calls) + 1
        self.status_calls.append({
            'status': status,
            'in_reply_to_id':in_reply_to_id,
            'media_ids':media_ids,
            'spoiler_text': spoiler_text,
            'quote_id': quote_id,
            'visibility': visibility,
            'id': current_call_id,
        })
        output_path = self._output_path / f"{current_call_id}.txt"
        with output_path.open("w", encoding="utf-8") as f:
            f.write(f"in_reply_to_id: {in_reply_to_id}\n")
            f.write(f"media_ids: {media_ids}\n")
            f.write("----------------------\n")
            f.write(status)

        return RtStatus(id=str(current_call_id), url=f"https://foo.com/{current_call_id}")

    def media_post(self, media_file: str, description:str) -> RtMedia:
        self.media_calls.append(media_file)
        current_id = len(self.media_calls)
        output_path = self._output_path / f"image_{current_id}.txt"

        filename = media_file.split("/")[-1]
        with output_path.open("w", encoding="utf-8") as f:
            f.write(f"id: {current_id}\n")
            f.write(f"media_file: {filename}\n")
            f.write(f"description: {description}\n")
        return RtMedia(id=str(current_id), file_path=Path(media_file))

    def get_auth_url(self) -> str:
        self.auth_url_call_count += 1
        return "https://foo.bar/blah"

    def log_in(self, code: str, to_file: str) -> None:
        self.login_calls.append([code, to_file])



from pathlib import Path
from typing import TypeVar

from robotooter.models.saveable_model import SaveableModel


class BotConfig(SaveableModel):
    bot_name: str
    bot_class: str
    username: str | None = None
    maintainer: str | None = None
    working_directory: Path
    tags: list[str] | None = None
    toot_prepend: str | None = None
    toot_append: str | None = None
    filter_names: list[str] = []

    model_config = {
        "extra": "allow"
    }


ConfigT = TypeVar('ConfigT', bound=BotConfig)


class RoboTooterConfig(SaveableModel):
    working_directory: Path
    plugin_filters: list[str] = []
    plugin_bots: list[str] = []



import os
from pathlib import Path
from typing import Any

from robotooter.bots.base_bot import BaseBot
from robotooter.filters.base_filter import BaseFilter
from robotooter.models.configs import BotConfig, RoboTooterConfig
from robotooter.util import load_included, load_list


class RoboTooter:
    __version__ = "0.2.0"
    def __init__(self, config: RoboTooterConfig) -> None:
        if config is None:
            config = RoboTooterConfig(
                working_directory=Path(os.path.expanduser("~/.robotooter")),
            )
        self.config = config
        self.filters: dict[str, type[BaseFilter]] = {}
        self.bots: dict[str, type[BaseBot]] = {}

        self._import_filters()
        self._import_bots()

    @property
    def is_configured(self) -> bool:
        return os.path.exists(self.config.working_directory / "config.json")

    def save_configuration(self) -> None:
        if not os.path.exists(self.config.working_directory):
            os.makedirs(self.config.working_directory)

        if not self.config.filename:
            self.config.filename = self.config.working_directory / "config.json"
        self.config.save()

    def create_new_bot(
        self,
        bot_name: str,
        username: str,
        maintainer: str,
        bot_class: str,
        filter_names: list[str],
        tags: list[str] | None = None,
        toot_prepend: str | None = None,
        toot_append: str | None = None,
    ) -> BaseBot:
        if bot_class not in self.bots.keys():
            raise ValueError(f"Invalid bot class {bot_class}")
        for filter_name in filter_names:
            if filter_name not in self.filters.keys():
                raise ValueError(f"Invalid filter name {filter_name}")

        bot_working_directory = self.config.working_directory / bot_name
        os.makedirs(bot_working_directory, exist_ok=True)

        if tags is None:
            tags = []

        if username[0] != '@':
            username = f"@{username}"
        if maintainer[0] != '@':
            maintainer = f"@{maintainer}"

        bot_config = BotConfig(
            bot_name=bot_name,
            bot_class=bot_class,
            filter_names=filter_names,
            username=username,
            working_directory=bot_working_directory,
            tags=tags,
            toot_prepend=toot_prepend,
            toot_append=toot_append,
            maintainer=maintainer,
        )
        self._save_bot_config(bot_config)
        new_bot = self.load_bot(bot_name)
        new_bot.install_package_resources()
        new_bot.post_creation_hook()
        return self.load_bot(bot_name)

    def load_bot(self, bot_name: str) -> BaseBot[Any]:
        bot_config = self._load_bot_config(bot_name)
        bot_class = self.bots[bot_config.bot_class]
        bot_filters: list[BaseFilter] = []
        for filter_name in bot_config.filter_names:
            bot_filters.append(self.filters[filter_name]())

        return bot_class.create_with_config(bot_config, bot_filters)

    def register_plugin(self, plugin_module: str) -> None:
        plugin_info = self._get_plugin_classes(plugin_module)
        if not plugin_info:
            raise RuntimeError(f"{plugin_module} is not a valid plugin module")
        if 'filters' in plugin_info:
            self.config.plugin_filters.extend(plugin_info['filters'])
        if 'bots' in plugin_info:
            self.config.plugin_bots.extend(plugin_info['bots'])

        self.save_configuration()

    def unregister_plugin(self, plugin_module: str) -> None:
        plugin_info = self._get_plugin_classes(plugin_module)
        if 'filters' in plugin_info:
            for filter_name in plugin_info['filters']:
                if filter_name in self.config.plugin_filters:
                    self.config.plugin_filters.remove(filter_name)
        if 'bots' in plugin_info:
            for bot in plugin_info['bots']:
                if bot in self.config.plugin_bots:
                    self.config.plugin_bots.remove(bot)
        self.save_configuration()

    def list_current_plugins(self) -> dict[str, list[str]]:
        return {
            'filters': self.config.plugin_filters,
            'bots': self.config.plugin_bots,
        }

    def list_all_bots(self) -> list[BotConfig]:
        configs = []
        for file in self.config.working_directory.iterdir():
            if file.is_dir():
                config_path = file / "config.json"
                if config_path.exists():
                    configs.append(BotConfig.load(config_path))
        return configs

    def _import_filters(self) -> None:
        self.filters = load_included('robotooter.filters', BaseFilter, 'Filter')
        self.filters = self.filters | load_list(self.config.plugin_filters, BaseFilter)

    def _import_bots(self) -> None:
        self.bots = load_included('robotooter.bots', BaseBot, 'Bot')
        self.bots = self.bots | load_list(self.config.plugin_bots, BaseBot)

    def _save_bot_config(self, bot_config: BotConfig) -> None:
        if not bot_config.filename:
            bot_config.filename = self.config.working_directory / f"{bot_config.bot_name}/config.json"
        bot_config.save()

    def _load_bot_config(self, bot_name: str) -> BotConfig:
        return BotConfig.load(self.config.working_directory / f"{bot_name}/config.json")

    def _get_plugin_classes(self, plugin_module: str) -> dict[str, str]:
        import importlib
        plugin_klass = importlib.import_module(plugin_module)
        info = {}
        if hasattr(plugin_klass, "plugin_filters"):
            info['filters'] = plugin_klass.plugin_filters
        if hasattr(plugin_klass, "plugin_bots"):
            info['bots'] = plugin_klass.plugin_bots
        return info

def load_robo_tooter(root_dir: str | None = '~/.robotooter') -> RoboTooter:
    if not root_dir:
        root_dir = '~/.robotooter'
    config_path = Path(os.path.expanduser(root_dir)) / "config.json"
    config = RoboTooterConfig.load(config_path)
    return RoboTooter(config)



from robotooter import RoboTooter


def run_configure(rt: RoboTooter) -> None:
    # Make sure it doesn't already exist
    if rt.is_configured:
        print("It looks like the configuration already exists. Exiting")
        return

    rt.save_configuration()
    print(f"Configuration saved to {rt.config.working_directory}")



from _typeshed import Incomplete

from robotooter import RoboTooter as RoboTooter
from robotooter.bots.base_bot import BaseBot as BaseBot
from robotooter.rt import load_robo_tooter as load_robo_tooter

parser: Incomplete
subparsers: Incomplete
plugins_parser: Incomplete
plugins_subparsers: Incomplete
register_parser: Incomplete
remove_parser: Incomplete
list_plugin_parser: Incomplete
list_parser: Incomplete

def require_bot(robo: RoboTooter, bot_name: str) -> BaseBot | None: ...
def find_bot_param(arguments: list[str]) -> tuple[str | None, list[str]]: ...
def find_working_dir(arguments: list[str]) -> tuple[str | None, list[str]]: ...
def find_param(arguments: list[str], flags: list[str]) -> tuple[str | None, list[str]]: ...
def main(arguments: list[str] | None = None) -> None: ...
def call_cli() -> None: ...



from pathlib import Path

from _typeshed import Incomplete
from mastodon import Mastodon

from robotooter.mastodon.mixin import MastodonMixin as MastodonMixin

class Authenticator(MastodonMixin):
    main_path: Incomplete
    bot_path: Incomplete
    full_username: Incomplete
    def __init__(self, main_path: Path, bot_path: Path, full_username: str) -> None: ...
    def get_auth_url(self) -> str: ...
    def log_in(self, code: str) -> None: ...
    @property
    def client(self) -> Mastodon: ...
    def get_or_create_app(self) -> Mastodon: ...
    def create_app(self) -> None: ...
    @property
    def server_config_path(self) -> Path: ...



from abc import ABC, abstractmethod


class BaseFilter(ABC):
    def process(self, line: str) -> str | None:
        """Override this"""
        return line

    def finalize(self) -> str | None:
        return ''

    @abstractmethod
    def reset(self) -> None:
        pass




from enum import StrEnum
from pathlib import Path

from pydantic import BaseModel


class RtMedia(BaseModel):
    id: str | None = None
    file_path: Path
    description: str | None = None

class StatusVisibility(StrEnum):
    Direct = 'direct' # post will be visible only to mentioned users, known in Mastodon’s UI as “Mentioned users only”
    Private = 'private' # post will be visible only to followers, known in Mastodon’s UI as “Followers only”
    Unlisted = 'unlisted' # post will be public but will not appear on the public timelines
    Public = 'public'
    # https://mastodonpy.readthedocs.io/en/stable/05_statuses.html#mastodon.Mastodon.status_post

class RtStatus(BaseModel):
    id: str| None = None
    text: str = ''
    in_reply_to_id: str | None = None
    url: str| None = None
    media: list[RtMedia] = []
    spoiler_text: str | None = None
    quote_id: str | None = None
    visibility: str | None = None



from pathlib import Path
from typing import Protocol

from _typeshed import Incomplete

RE_BOTLESS: Incomplete
RE_MODNAME: Incomplete

class HasWorkingDirectory(Protocol):
    working_directory: Path
    def pkg_resources_path(self) -> Path: ...
    def resource_path(self, sub_path: str | None = None) -> Path: ...

class ResourcesMixin:
    def pkg_resources_path(self) -> Path: ...
    def install_package_resources(self) -> None: ...
    def resource_path(self, sub_path: str | None = None) -> Path: ...



# Copyright (C) 2024 Bryan L. Fordham
#
# This file is part of RoboTooter.
#
# RoboTooter is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RoboTooter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with RoboTooter. If not, see <https://www.gnu.org/licenses/>.

import argparse
import sys

from robotooter import RoboTooter
from robotooter.bots.base_bot import BaseBot
from robotooter.rt import load_robo_tooter

parser = argparse.ArgumentParser()
parser.add_argument('-b', '--bot')
parser.add_argument('-w', '--working-dir', dest='working_dir', default='~/.robotooter')

subparsers = parser.add_subparsers(help='subcommand help', dest='command', required=False)

subparsers.add_parser('info', help='Display info about the local setup')
subparsers.add_parser('configure', help='Configure the local setup')
subparsers.add_parser('version', help='Display the current version')
subparsers.add_parser('help', help='Display this help message')
subparsers.add_parser('list', help='List current bots')
subparsers.add_parser('create', help='Create a new bot')

plugins_parser = subparsers.add_parser('plugins', help='Manage plugins')
plugins_subparsers = plugins_parser.add_subparsers(
    dest='plugin_action',
    help='Plugin management commands'
)
register_parser = plugins_subparsers.add_parser('register', help='Register a new plugin')
register_parser.add_argument('-p', '--plugin', help='Name of the plugin to register')

remove_parser = plugins_subparsers.add_parser('remove', help='Remove a plugin')
remove_parser.add_argument('-p', '--plugin', help='Name of the plugin to remove')

list_plugin_parser = plugins_subparsers.add_parser('list', help='List plugins')

list_parser = plugins_subparsers.add_parser('help', help='Print this help message')

def require_bot(robo: RoboTooter, bot_name:str) -> BaseBot|None:
    return robo.load_bot(bot_name)


def find_bot_param(arguments: list[str]) -> tuple[str | None, list[str]]:
    return find_param(arguments, ['-b', '--bot'])


def find_working_dir(arguments: list[str]) -> tuple[str | None, list[str]]:
    return find_param(arguments, ['-w', '--working-dir'])


def find_param(arguments: list[str], flags: list[str]) -> tuple[str | None, list[str]]:
    i = -1
    flag_value = None
    for p in flags:
        if p in arguments:
            i = arguments.index(p)
        if i > -1:
            break
    if i > -1:
        flag_value = arguments[i+1]
        del arguments[i+1]
        del arguments[i]
    return flag_value, arguments


def main(arguments: list[str] | None = None) -> None:
    if arguments:
        bot_name, arguments = find_bot_param(arguments)
        if bot_name:
            wd, arguments = find_working_dir(arguments)
            rt = load_robo_tooter(wd)
            bot = require_bot(rt, bot_name)
            if not bot:
                print(f"{bot_name} is not a valid bot")
                sys.exit(1)
            bot.parse_and_run(arguments)
            return

    args, remainder = parser.parse_known_args(arguments)
    rt = load_robo_tooter(args.working_dir)

    match args.command:
        case 'info':
            from robotooter.cli.info import run_info
            run_info(rt)

        case 'configure':
            from robotooter.cli.configure import run_configure
            run_configure(rt)

        case 'version':
            print(rt.__version__)

        case 'help':
            parser.print_help()

        case 'create':
            from robotooter.cli.create_bot import run_create_bot
            run_create_bot(rt)

        case 'list':
            all_plugins = rt.list_all_bots()
            if not all_plugins:
                print("No plugins found")
            if not all_plugins:
                print("No bots are currently configured")
                return
            for b_config in all_plugins:
                if not b_config:
                    continue
                print(b_config.bot_name)
                print(f"  {b_config.username}")
                print(f"  {b_config.bot_class}")

        case 'plugins':
            match args.plugin_action:
                case 'register':
                    rt.register_plugin(args.plugin)
                    print(f"{args.plugin} plugin registered")
                case 'remove':
                    rt.unregister_plugin(args.plugin)
                    print(f"{args.plugin} plugin removed")
                case 'list':
                    plugin_info = rt.list_current_plugins()
                    if plugin_info['filters']:
                        print("Plugin filters:")
                        for f in plugin_info['filters']:
                            print(f)
                    if plugin_info['bots']:
                        print("Plugin bots:")
                        for b in plugin_info['bots']:
                            print(b)
                case 'help':
                    parser.print_help()
                case _:
                    parser.print_help()


        case _:
            parser.print_help()

def call_cli() -> None:
    main(sys.argv[1:])

if __name__ == '__main__':
    call_cli()



from pathlib import Path
from typing import Self

from pydantic import BaseModel, Field


class SaveableModel(BaseModel):
    filename: Path | None = Field(exclude=True, default=None)

    @classmethod
    def load(cls, filename: Path) -> Self:
        data = filename.read_text()
        instance = cls.model_validate_json(data)
        instance.filename = filename
        return instance

    def save(self) -> None:
        if self.filename:
            self.filename.write_text(self.model_dump_json(indent=2))



import argparse

from robotooter.bots.base_bot import COUNT_PARAM as COUNT_PARAM
from robotooter.bots.base_bot import BaseBot as BaseBot
from robotooter.bots.base_bot import command as command
from robotooter.filters.base_filter import BaseFilter as BaseFilter
from robotooter.models.configs import BotConfig as BotConfig
from robotooter.models.configs import ConfigT as ConfigT

class FakeBotConfig(BotConfig):
    extra_field_1: str

class FakeBot(BaseBot[FakeBotConfig]):
    NAME: str
    DESCRIPTION: str
    CONFIG_CLASS = FakeBotConfig
    @classmethod
    def create_with_config(cls, config_data: BotConfig, filters: list[BaseFilter]) -> FakeBot: ...
    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None: ...
    number: int
    def __init__(self, config: FakeBotConfig, filters: list[BaseFilter]) -> None: ...
    def add_one(self, args: argparse.Namespace) -> None: ...
    def set_number(self, args: argparse.Namespace) -> None: ...
    def _setup_data(self, args: argparse.Namespace) -> None: ...
    def _toot(self, args: argparse.Namespace) -> None: ...



from robotooter.filters.base_filter import BaseFilter as BaseFilter

PG_START: str
PG_END: str

class GutenbergFilter(BaseFilter):
    def __init__(self) -> None: ...
    def process(self, line: str) -> str | None: ...
    def reset(self) -> None: ...




from mastodon import Mastodon
from mastodon.return_types import Status

from robotooter.mastodon.mixin import MastodonMixin
from robotooter.models.configs import BotConfig
from robotooter.models.statuses import RtMedia, RtStatus


class MastodonManager(MastodonMixin):
    def __init__(self, bot_config: BotConfig) -> None:
        if not bot_config.username:
            raise RuntimeError("Username is not set")

        self.bot_config = bot_config
        self.bot_path = bot_config.working_directory
        self._client: Mastodon | None = None
        self.username, self.server = self.split_username(bot_config.username)

    @property
    def client(self) -> Mastodon:
        if self._client is None:
            self._client = Mastodon(access_token=str(self.bot_config_path))

        return self._client

    def toot(self, status: RtStatus | str) -> RtStatus:
        if isinstance(status, str):
            status = RtStatus(text=status)

        if len(status.media):
            for i, media in enumerate(status.media):
                if not media.id:
                    status.media[i] = self.post_media(media)
        media_ids = [m.id for m in status.media]

        m_status = self.client.status_post(
            status = status.text,
            in_reply_to_id=status.in_reply_to_id,
            media_ids=media_ids,
            spoiler_text=status.spoiler_text,
            quote_id=status.quote_id,
            visibility=status.visibility,
        )
        return self._adapt_status(status, m_status)

    def thread(self, toots: list[RtStatus]) -> None:
        in_reply_to_id = None
        for i, toot in enumerate(toots):
            toot.in_reply_to_id = in_reply_to_id
            status = self.toot(toot)
            toots[i] = status
            in_reply_to_id = status.id

    def post_media(self, media: RtMedia) -> RtMedia:
        m_media = self.client.media_post(
            media_file=str(media.file_path),
            description=media.description,
        )
        media.id = m_media.id
        return media

    def _adapt_status(self, rt_status: RtStatus, m_status: Status) -> RtStatus:
        rt_status.id = m_status.id
        rt_status.url = m_status.url
        return rt_status




from typing import Any

from _typeshed import Incomplete

from robotooter.bots.base_bot import BaseBot as BaseBot
from robotooter.filters.base_filter import BaseFilter as BaseFilter
from robotooter.models.configs import BotConfig as BotConfig
from robotooter.models.configs import RoboTooterConfig as RoboTooterConfig
from robotooter.util import load_included as load_included
from robotooter.util import load_list as load_list

class RoboTooter:
    __version__: str
    config: Incomplete
    filters: dict[str, type[BaseFilter]]
    bots: dict[str, type[BaseBot]]
    def __init__(self, config: RoboTooterConfig) -> None: ...
    @property
    def is_configured(self) -> bool: ...
    def save_configuration(self) -> None: ...
    def create_new_bot(self,
        bot_name: str,
        username: str,
        maintainer: str,
        bot_class: str,
        filter_names: list[str],
        tags: list[str] | None = None,
        toot_prepend: str | None = None,
        toot_append: str | None = None
    ) -> BaseBot: ...
    def load_bot(self, bot_name: str) -> BaseBot[Any]: ...
    def register_plugin(self, plugin_module: str) -> None: ...
    def unregister_plugin(self, plugin_module: str) -> None: ...
    def list_current_plugins(self) -> dict[str, list[str]]: ...
    def list_all_bots(self) -> list[BotConfig]: ...

def load_robo_tooter(root_dir: str | None = '~/.robotooter') -> RoboTooter: ...



from pathlib import Path
from typing import TypeVar

from robotooter.models.saveable_model import SaveableModel as SaveableModel

class BotConfig(SaveableModel):
    bot_name: str
    bot_class: str
    username: str | None
    maintainer: str | None
    working_directory: Path
    tags: list[str] | None
    toot_prepend: str | None
    toot_append: str | None
    filter_names: list[str]
    class Config:
        extra: str
ConfigT = TypeVar('ConfigT', bound=BotConfig)

class RoboTooterConfig(SaveableModel):
    working_directory: Path
    plugin_filters: list[str]
    plugin_bots: list[str]



from robotooter import RoboTooter as RoboTooter
from robotooter.bots.base_bot import BaseBot as BaseBot
from robotooter.cli.util import get_yes as get_yes
from robotooter.mastodon.authenticator import Authenticator as Authenticator

def run_authorize(rt: RoboTooter, bot: BaseBot, force: bool = False) -> None: ...



from pathlib import Path
from typing import IO, Any, Iterable, Sequence

from _typeshed import Incomplete

from robotooter.filters.base_filter import BaseFilter as BaseFilter

FileOrPath = str | Path | IO[Any]

class FileFilter:
    filters: Incomplete
    def __init__(self, filters: list[BaseFilter] | None = None) -> None: ...
    def process_files(self, input_files: Sequence[FileOrPath], output_path_or_file: FileOrPath) -> None: ...
    def process_file(self, input_path_or_file: FileOrPath, output_path_or_file: FileOrPath | None) -> None: ...
    def filter_lines(self, lines: Iterable[str]) -> Iterable[str]: ...



import argparse
import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Callable, ClassVar, Generic, Protocol, Type, TypeVar

from robotooter.filters.base_filter import BaseFilter
from robotooter.mastodon.manager import MastodonManager
from robotooter.models.configs import BotConfig, ConfigT


class CommandHandler(Protocol):
    _command_name: str
    _command_help: str
    _command_params: list[tuple[str, str, dict[str, Any]]] | None

    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...


F = TypeVar('F', bound=Callable[..., Any])
CommandParam = tuple[str, str, dict[str, Any]]
CommandRegistry = dict[str, list[CommandParam]]


def command(
        name: str,
        help_text: str | None = None,
        params: list[CommandParam] | None = None
) -> Callable[[F], F]:
    """Decorator to register a method as a command handler with help text."""
    def decorator(func: F) -> F:
        setattr(func, '_command_name', name)
        setattr(func, '_command_help', help_text or func.__doc__ or f"Execute {name} command")
        setattr(func, '_command_params', params or [])
        return func
    return decorator


COUNT_PARAM: CommandParam = ('-c', '--count', {'default': 1, 'type': int, 'help': 'Quantity'})


class BaseBot(ABC, Generic[ConfigT]):
    NAME = "BaseTooter"
    DESCRIPTION = "BaseTooter bot"

    CONFIG_CLASS: ClassVar[Type[BotConfig]] = BotConfig

    @classmethod
    @abstractmethod
    def create_with_config(cls, config_data: ConfigT, filters: list[BaseFilter]) -> 'BaseBot[Any]':
        pass

    @staticmethod
    def new_bot_info() -> str | None:
        return None

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        """Override this to add bot-specific arguments."""
        pass

    def __init__(self, config: ConfigT, filters: list[BaseFilter]) -> None:
        self.config = config
        self.config.filename = config.filename
        self.working_directory: Path = config.working_directory
        self.data_root = Path(os.path.join(self.working_directory, "data"))
        self.plugin_filters = filters
        self.mastodon_manager = MastodonManager(self.config)

        self.command_handlers: dict[str, CommandHandler] = {}
        self.command_overrides: CommandRegistry = {}

    @command('setup', 'Initialize bot configuration')
    def setup_data(self, args: argparse.Namespace) -> None:
        self._setup_data(args)

    @command('toot', 'Send a toot', params=[COUNT_PARAM])
    def toot(self, args: argparse.Namespace) -> None:
        self._toot(args)

    @abstractmethod
    def _setup_data(self, args: argparse.Namespace) -> None:
        pass

    @abstractmethod
    def _toot(self, args: argparse.Namespace) -> None:
        pass

    def install_package_resources(self) -> None:
        pass

    def parse_and_run(self, args_list: list[str]) -> None:
        self._register_commands()
        parser = argparse.ArgumentParser(prog=f'robotooter -b {self.config.bot_name}')

        # Add bot-specific arguments
        self.add_arguments(parser)

        # Add subparsers for commands
        subparsers = parser.add_subparsers(dest='command', help='Available commands')
        for cmd_name, handler in self.command_handlers.items():
            help_text = getattr(handler, '_command_help', f'Execute {cmd_name}')
            sub_parser = subparsers.add_parser(cmd_name, help=help_text)
            if cmd_name in self.command_overrides:
                params = self.command_overrides[cmd_name]
            else:
                params = getattr(handler, '_command_params')
            if params:
                for short, long, attrs in params:
                    sub_parser.add_argument(short, long, **attrs)


        args, remaimder = parser.parse_known_args(args_list)

        if args.command in self.command_handlers:
            self.command_handlers[args.command](args)
            return

        parser.print_help()

    def post_creation_hook(self) -> None:
        pass

    def _register_commands(self) -> None:
        if not self.command_handlers:
            for attr_name in dir(self):
                attr = getattr(self, attr_name)
                if hasattr(attr, '_command_name'):
                    self.command_handlers[attr._command_name] = attr




from pathlib import Path

from mastodon import Mastodon

from robotooter.mastodon.mixin import MastodonMixin


class Authenticator(MastodonMixin):
    def __init__(self, main_path: Path, bot_path: Path, full_username: str) -> None:
        self.main_path = main_path
        self.bot_path = bot_path
        self.full_username = full_username

        self._client: Mastodon | None = None

        self.username, self.server = self.split_username(full_username)

    def get_auth_url(self) -> str:
        return self.client.auth_request_url()

    def log_in(self, code: str) -> None:
        self.client.log_in(
            code=code,
            to_file=str(self.bot_config_path)
        )

    @property
    def client(self) -> Mastodon:
        if self._client is None:
            self._client = self.get_or_create_app()

        return self._client

    def get_or_create_app(self) -> Mastodon:
        app_config_path = self.server_config_path
        app_config_path.parent.mkdir(parents=True, exist_ok=True)

        if not app_config_path.exists():
            self.create_app()

        return Mastodon(str(app_config_path))

    def create_app(self) -> None:
        api_base_url = 'https://' + self.server
        Mastodon.create_app(
            'RoboTooter',
            to_file=str(self.server_config_path),
            api_base_url=api_base_url,
        )

    @property
    def server_config_path(self) -> Path:
        return self.main_path / "apps" / f"{self.server}.app"




from typing import List

from robotooter.filters.base_filter import BaseFilter

USE_NL = ['CHAPTER ', 'PART ']


class ParagraphCombiningFilter(BaseFilter):
    def __init__(self) -> None:
        self._buffer: List[str] = []

    def reset(self) -> None:
        self._buffer = []

    def process(self, line: str) -> str | None:
        stripped = line.strip()
        if stripped == "":
            if self._buffer:
                return self._join_buffer()
            return None
        else:
            self._buffer.append(stripped)
            return None  # Buffering, not yielding yet

    def finalize(self) -> str | None:
        if self._buffer:
            return self._join_buffer()
        return None

    def _join_buffer(self) -> str:
        first_line = self._buffer[0]
        joiner = ' '
        for nl in USE_NL:
            if first_line.startswith(nl):
                joiner = '\n'

        output = joiner.join(self._buffer) + '\n\n'
        self._buffer.clear()
        return output




from robotooter import RoboTooter


def run_info(rt: RoboTooter) -> None:
    print("RobotoTooter info")
    print(f"Version: {rt.__version__}")
    print(f"Root directory: {rt.config.working_directory}")

    if not rt.is_configured:
        print("It looks like things have not been set up yet.")
        return

    print("Filters:")
    for f in rt.filters.keys():
        print(f"  {f}")
    print("Tooters:")
    for t in rt.bots:
        print(f"  {t}")




from robotooter import RoboTooter as RoboTooter
from robotooter.cli.util import get_yes as get_yes

def get_options(text: str, options: dict[str, type], required: bool = False, only_one: bool = False) -> list[str]: ...
def run_create_bot(rt: RoboTooter) -> None: ...



from robotooter.filters.base_filter import BaseFilter

PG_START = "*** START OF THE PROJECT GUTENBERG EBOOK"
PG_END = "*** END OF THE PROJECT GUTENBERG EBOOK"


class GutenbergFilter(BaseFilter):
    """
    For this we read in each file until we find the end of the Project Gutenberg header.
    Then we keep reading, and write to our combined file. We stop when we're back to boilerplace.
    """
    def __init__(self) -> None:
        self._write_file = False
        self._in_note = False

    def process(self, line: str) -> str | None:
        if self._should_send(line):
            return line

        return None

    def reset(self) -> None:
        self._write_file = False
        self._in_note = False

    def _should_send(self, line: str) -> bool:
        line = line.strip()

        if not line:
            return self._write_file

        if line.find(PG_END) > -1:
            self._write_file = False
            return False

        if line.find('*') > -1:
            pass

        if self._write_file:
            if line[0] == '*':
                return False

            if line.find('[') > -1:
                self._in_note = True

            if line.find(']') > -1:
                self._in_note = False
                return False

        if self._write_file and not self._in_note:
            return True

        if line.find(PG_START) > -1:
            self._write_file = True

        return False



from pathlib import Path
from typing import Self

from pydantic import BaseModel

class SaveableModel(BaseModel):
    filename: Path | None
    @classmethod
    def load(cls, filename: Path) -> Self: ...
    def save(self) -> None: ...



from abc import ABCMeta
from pathlib import Path
from typing import TypeVar

T = TypeVar('T')

def load_included(package_path: str, expected_klass: type, ending: str) -> dict[str, type]: ...
def load_list(names: list[str], expected_type: type[T] | ABCMeta) -> dict[str, type[T]]: ...
def download_sources(source_file_path: Path, output_directory: Path) -> None: ...



from _typeshed import Incomplete

from robotooter.filters.base_filter import BaseFilter as BaseFilter

USE_NL: Incomplete

class ParagraphCombiningFilter(BaseFilter):
    def __init__(self) -> None: ...
    def reset(self) -> None: ...
    def process(self, line: str) -> str | None: ...
    def finalize(self) -> str | None: ...



from typing import Type

from robotooter import RoboTooter
from robotooter.cli.util import get_yes


def get_options(text:str, options: dict[str, Type], required: bool = False, only_one: bool = False)-> list[str]:
    keys = sorted(options.keys())

    for i, key in enumerate(keys):
        print(f"{i+1}. {key}")

    choices = input("Select an option: ")
    parts = choices.split(",")
    chosen_keys = [keys[int(x.strip()) - 1] for x in parts if x.strip()]
    if not chosen_keys:
        if required:
            print("\nsorry, you must choose at least one option\n")
            return get_options(text, options, required=required, only_one=only_one)
    if only_one and len(chosen_keys) > 1:
        print("\nsorry, you must choose ONLY one option\n")
        return get_options(text, options, required=required, only_one=only_one)

    return chosen_keys


def run_create_bot(rt: RoboTooter) -> None:
    bot_name = ''
    username = ''
    bot_class = ''
    filter_names: list[str] = []
    maintainer = ''

    all_set = False
    while not all_set:
        print("Let's get your bot's information.")
        bot_name = input("Enter the bot name: ")
        username = input("Enter the username, such as 'mybot@mastodon.social': ")
        bot_class = get_options("Select a bot class", rt.bots, True, True)[0]
        filter_names = get_options("Select filters", rt.filters)
        maintainer = input("Enter the username of the maintainer, such as 'mybot@mastodon.social': ")

        print("\nHere's what we have:")
        print("bot_name:", bot_name)
        print("username:", username)
        print("bot_class:", bot_class)
        print("filter_names:", filter_names)
        print("maintainer:", maintainer)
        print("")
        all_set = get_yes("Is this correct?")

    bot = rt.create_new_bot(
        bot_name=bot_name,
        username=username,
        bot_class=bot_class,
        filter_names=filter_names,
        maintainer=maintainer,
    )
    print("\nBot created successfully")

    if get_yes("Do you want to go ahead and authorize with Mastodon?"):
        from robotooter.cli.authorize import run_authorize
        run_authorize(rt, bot)
    else:
        print("OK, you can run the authorize command later.")

    if bot.new_bot_info():
        print(bot.new_bot_info())





import abc
import argparse
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Callable, ClassVar, Generic, Protocol, TypeVar

from _typeshed import Incomplete

from robotooter.filters.base_filter import BaseFilter as BaseFilter
from robotooter.mastodon.manager import MastodonManager as MastodonManager
from robotooter.models.configs import BotConfig as BotConfig
from robotooter.models.configs import ConfigT as ConfigT

class CommandHandler(Protocol):
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...
F = TypeVar('F', bound=Callable[..., Any])
CommandParam = tuple[str, str, dict[str, Any]]
CommandRegistry = dict[str, list[CommandParam]]

def command(name: str, help_text: str | None = None, params: list[CommandParam] | None = None) -> Callable[[F], F]: ...

COUNT_PARAM: CommandParam

class BaseBot(ABC, Generic[ConfigT], metaclass=abc.ABCMeta):
    NAME: str
    DESCRIPTION: str
    CONFIG_CLASS: ClassVar[type[BotConfig]]
    @classmethod
    @abstractmethod
    def create_with_config(cls, config_data: ConfigT, filters: list[BaseFilter]) -> BaseBot[Any]: ...
    @staticmethod
    def new_bot_info() -> str | None: ...
    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None: ...
    config: Incomplete
    working_directory: Path
    data_root: Incomplete
    plugin_filters: Incomplete
    mastodon_manager: Incomplete
    command_handlers: dict[str, CommandHandler]
    command_overrides: CommandRegistry
    def __init__(self, config: ConfigT, filters: list[BaseFilter]) -> None: ...
    def setup_data(self, args: argparse.Namespace) -> None: ...
    def toot(self, args: argparse.Namespace) -> None: ...
    def install_package_resources(self) -> None: ...
    def parse_and_run(self, args_list: list[str]) -> None: ...
    def post_creation_hook(self) -> None: ...



from _typeshed import Incomplete
from mastodon import Mastodon
from mastodon.return_types import Status as Status

from robotooter.mastodon.mixin import MastodonMixin as MastodonMixin
from robotooter.models.configs import BotConfig as BotConfig
from robotooter.models.statuses import RtMedia as RtMedia
from robotooter.models.statuses import RtStatus as RtStatus

class MastodonManager(MastodonMixin):
    bot_config: Incomplete
    bot_path: Incomplete
    def __init__(self, bot_config: BotConfig) -> None: ...
    @property
    def client(self) -> Mastodon: ...
    def toot(self, status: RtStatus | str) -> RtStatus: ...
    def thread(self, toots: list[RtStatus]) -> None: ...
    def post_media(self, media: RtMedia) -> RtMedia: ...



import os
from argparse import Namespace

import markovify

from robotooter.bots.base_bot import BaseBot
from robotooter.bots.markov.markov_processor import MarkovProcessor
from robotooter.filters.base_filter import BaseFilter
from robotooter.models.configs import BotConfig
from robotooter.util import download_sources


class MarkovBot(BaseBot[BotConfig]):
    NAME = "MarkovBot"

    @classmethod
    def create_with_config(cls, config_data: BotConfig, filters: list[BaseFilter]) -> 'MarkovBot':
        return cls(config_data, filters)

    @staticmethod
    def new_bot_info() -> str | None:
        return "You will need to add `sources.txt` to the working directory"

    def __init__(self, config: BotConfig, filters: list[BaseFilter]) -> None:
        super().__init__(config, filters)
        self.source_dir = self.data_root / "sources"
        self.consolidated_file_path = self.data_root / "consolidated.txt"
        self.model_path = self.data_root / "model.json"
        self._model: markovify.Text | None = None

    def _toot(self, args: Namespace) -> None:
        for _ in range(args.count):
            for text in self.generate_toot():
                self.mastodon_manager.toot(text)

    def generate_toot(self) -> list[str]:
        content = None
        while not content:  # Sometimes we get None, prevent that
            content = self.model.make_sentence()
        tags = ''
        if self.config.tags:
            tags = "#" + " #".join(self.config.tags)
        return [f"{content}\n\n{tags.strip()}"]

    def _setup_data(self, args: Namespace) -> None:
        if self.model_exists:
            raise Exception(f"Model file {self.model_path} already exists.")

        download_sources(self.working_directory / "sources.txt", self.source_dir)
        self.processor.preprocess_sources()
        self.processor.build_model()

    @property
    def processor(self) -> MarkovProcessor:
        return MarkovProcessor(self.working_directory, self.plugin_filters)

    @property
    def model_exists(self) -> bool:
        return os.path.exists(self.model_path)

    @property
    def model(self) -> markovify.Text:
        if self._model is None:
            json_blob = open(self.model_path, "r").read()
            self._model = markovify.Text.from_json(json_blob)

        return self._model




from pathlib import Path

class MastodonMixin:
    def split_username(self, full_username: str) -> tuple[str, str]: ...
    @property
    def bot_config_exists(self) -> bool: ...
    @property
    def bot_config_path(self) -> Path: ...



from io import TextIOWrapper
from pathlib import Path
from typing import IO, Any, Iterable, Sequence

from robotooter.filters.base_filter import BaseFilter

FileOrPath = str | Path | IO[Any]

class FileFilter:
    def __init__(self, filters: list[BaseFilter]|None = None):
        self.filters = filters or []

    def process_files(self, input_files: Sequence[FileOrPath], output_path_or_file: FileOrPath)-> None:
        output_file, output_opened = self._open(output_path_or_file, 'w')
        try:
            for file in input_files:
                self.process_file(file, output_file)
        except Exception as e:
            print(e)
            raise e
        finally:
            if output_opened and output_file:
                output_file.close()

    def process_file(self, input_path_or_file: FileOrPath, output_path_or_file: FileOrPath | None) -> None:
        input_file, input_opened  = self._open(input_path_or_file)
        if input_file is None:
            raise FileNotFoundError("Can't read None, yo")
        output_file, output_opened = self._open(output_path_or_file, 'w')
        [f.reset() for f in self.filters]

        try:
            for line in self.filter_lines(input_file):
                if output_file:
                    output_file.write(line)
        except Exception as e:
            print(e)
            raise e
        finally:
            if input_opened:
                input_file.close()
            if output_opened and output_file:
                output_file.close()

    def filter_lines(self, lines: Iterable[str]) -> Iterable[str]:
        for line in lines:
            for f in self.filters:
                result = f.process(line)
                if result is None:
                    break  # Drop line
                line = result  # Pass updated line to next filter
            else:
                yield line  # Yield only if not dropped by any filter

    def _open(self, file_or_path: FileOrPath | None, mode:str='r') -> tuple[IO|None, bool]:
        if file_or_path is None:
            return None, False

        if isinstance(file_or_path, TextIOWrapper):
            return file_or_path, False

        return open(str(file_or_path), mode), True



import glob
import os
import sys
from pathlib import Path

import markovify

from robotooter.file_filter import FileFilter
from robotooter.filters.base_filter import BaseFilter

DEFAULT_STATE_SIZE = 3

class MarkovProcessor:
    def __init__(self, working_directory: Path, filters: list[BaseFilter]):
        self.filters = filters
        self.working_directory = working_directory
        self.data_dir = working_directory / "data"

        self.source_dir = self.data_dir / "sources"
        self.consolidated_file_path = self.data_dir / "consolidated.txt"
        self.model_path = self.data_dir / "model.json"
        self._model: markovify.Text | None = None

    def preprocess_sources(self) -> None:
        input_files = glob.glob(os.path.join(self.source_dir, "*.txt"))
        if not input_files:
            print("No input files found. Cowardly refusing to go on.")
            sys.exit(1)

        input_files.sort()

        processor = FileFilter(filters=self.filters)
        processor.process_files(input_files, self.consolidated_file_path)

    def build_model(self, state_size: int = DEFAULT_STATE_SIZE) -> None:
        if not os.path.exists(self.consolidated_file_path):
            raise Exception("Consolidated file not found. Cowardly refusing to go on.")

        all_data = open(self.consolidated_file_path, "r").read()
        text_model = markovify.Text(all_data, state_size=state_size)

        with open(self.model_path, "w") as model_file:
            model_file.write(text_model.to_json())

    @property
    def consolidated_exists(self) -> bool:
        return os.path.exists(self.consolidated_file_path)

    @property
    def model_exists(self) -> bool:
        return os.path.exists(self.model_path)



from argparse import Namespace
from typing import Any

import markovify
from _typeshed import Incomplete

from robotooter.bots.base_bot import BaseBot as BaseBot
from robotooter.bots.markov.markov_processor import MarkovProcessor as MarkovProcessor
from robotooter.filters.base_filter import BaseFilter as BaseFilter
from robotooter.models.configs import BotConfig as BotConfig
from robotooter.models.configs import ConfigT
from robotooter.util import download_sources as download_sources

class MarkovBot(BaseBot):
    @classmethod
    def create_with_config(cls, config_data: ConfigT, filters: list[BaseFilter]) -> 'BaseBot[Any]': ...

    NAME: str
    @staticmethod
    def new_bot_info() -> str | None: ...
    source_dir: Incomplete
    consolidated_file_path: Incomplete
    model_path: Incomplete
    def __init__(self, config: BotConfig, filters: list[BaseFilter]) -> None: ...
    def generate_toot(self) -> list[str]: ...
    @property
    def processor(self) -> MarkovProcessor: ...
    @property
    def model_exists(self) -> bool: ...
    @property
    def model(self) -> markovify.Text: ...
    def _setup_data(self, args: Namespace) -> None: ...
    def _toot(self, args: Namespace) -> None: ...
